#include <stdio.h>
#include <dirent.h>
#include <vector>
#include <string>
#include <algorithm>
#include <libexif/exif-data.h>
#include <libexif/exif-loader.h>
#include <libexif/exif-mem.h>
#include <libexif/exif-ifd.h>
#include <libexif/exif-data-type.h>
#include <libexif/exif-content.h>
#include <libexif/exif-format.h>
using namespace std;

void usage() {
    printf("Usage: addgeoref folder\n");
    printf("  folder should contain a file called data.txt and the images to process\n");
}

struct latlon {
    long lat;
    long lon;
};

static ExifEntry *create_tag(ExifData *exif, ExifIfd ifd, ExifTag tag, size_t len, int components, ExifFormat format)
{
    void *buf;
    ExifEntry *entry;

    /* Create a memory allocator to manage this ExifEntry */
    ExifMem *mem = exif_mem_new_default();

    /* Create a new ExifEntry using our allocator */
    entry = exif_entry_new_mem (mem);

    /* Allocate memory to use for holding the tag data */
    buf = exif_mem_alloc(mem, len);

    /* Fill in the entry */
    entry->data = (unsigned char*)buf;
    entry->size = len;
    entry->tag = tag;
    entry->components = components;
    entry->format = format;//EXIF_FORMAT_UNDEFINED;

    /* Attach the ExifEntry to an IFD */
    exif_content_add_entry (exif->ifd[ifd], entry);

    /* The ExifMem and ExifEntry are now owned elsewhere */
    exif_mem_unref(mem);
    exif_entry_unref(entry);

    return entry;
}

int main(int argc, char** argv) {

    if ( argc < 2 ) {
        usage();
        return 0;
    }

    char* inputFolder = argv[1];

    char datafile[128];
    sprintf(datafile, "%s/DATA.TXT", inputFolder);
    FILE* f = fopen(datafile, "r");
    if ( !f ) {
        printf("Could not find DATA.TXT\n");
        return -2;
    }

    printf("Found DATA.TXT\n");

    char * line = NULL;
    size_t len = 0;
    vector<latlon> geopoints;
    while (getline(&line, &len, f) != -1) {
        string s = string(line);
        size_t ind = s.find(",");
        if ( ind == string::npos ) {
            printf("Data file should contain lat/lon separated by commas on each line\n");
            fclose(f);
            return -1;
        }
        string latstring = s.substr(0, ind);
        string lonstring = s.substr(ind+1);
        //printf("   %s,   %s", latstring.c_str(), lonstring.c_str());
        latlon ll;
        ll.lat = atol( latstring.c_str() );
        ll.lon = atol( lonstring.c_str() );
        geopoints.push_back(ll);
        free(line);
        line = NULL;
        len = 0;
    }

    fclose(f);

    //for (int i = 0; i < geopoints.size(); i++)
    //    printf(" ll  %ld,   %ld \n",geopoints[i].lat, geopoints[i].lon);

    printf("Data file contains %d points.\n", geopoints.size());

    DIR* dp = opendir(inputFolder);
    if ( ! dp ) {
        printf("Could not open directory: %s\n", inputFolder);
        return -1;
    }

    struct dirent* ep;
    vector<string> jpegs;
    while (ep = readdir(dp)) {
        string s = ep->d_name;
        if (s.find("jpg") != string::npos ||
                s.find("jpeg") != string::npos ||
                s.find("JPG") != string::npos ||
                s.find("JPEG") != string::npos)
            jpegs.push_back(ep->d_name);
    }

    closedir(dp);

    sort(jpegs.begin(), jpegs.end());

    //for (int i = 0; i < jpegs.size(); i++)
    //    printf("   %s\n",jpegs[i].c_str());

    printf("Found %d JPEG files.\n", jpegs.size());

    if ( geopoints.size() != jpegs.size() ) {
        printf("The number of JPEG files should match the number of location points.\n");
        return -1;
    }

    for (int i = 0; i < jpegs.size(); i++) {

        string filename = string(inputFolder)+"/"+jpegs[i];
        string outfilename = string(inputFolder)+"/out_"+jpegs[i];
        latlon &ll = geopoints[i];

        printf("Processing %s:   %ld, %ld\n",filename.c_str(), ll.lat, ll.lon);

        FILE* fin = fopen(filename.c_str(), "rb");
        char header[2];
        fread(header,2,1,fin); // FF D8
        fread(header,2,1,fin); // FF E1

        unsigned short blockLen = 0;
        char* swizzle = (char*)&blockLen;

        fread(&swizzle[1],1,1,fin);
        fread(&swizzle[0],1,1,fin);

        //printf("Skip to %d\n", blockLen+4);

        fseek(fin, blockLen+4, SEEK_SET);

        ExifData* data = exif_data_new_from_file(filename.c_str());

        //exif_data_dump(data);

        ExifEntry* cameraModelEntry = exif_data_get_entry(data, exif_tag_from_name("Model"));
        //exif_entry_dump(cameraModelEntry, 4);
        //printf("Camera model is %s\n",cameraModelEntry->data);

        bool yi4k = string((char*)cameraModelEntry->data) == "YDXJ 2";

        //ExifEntry* versionIdEntry = NULL;//exif_data_get_entry(data, exif_tag_from_name("GPSVersionID"));
        ExifEntry* latitudeEntry = yi4k ? exif_data_get_entry(data, exif_tag_from_name("GPSLatitude")) : NULL;
        ExifEntry* longitudeEntry = yi4k ? exif_data_get_entry(data, exif_tag_from_name("GPSLongitude")) : NULL;
        ExifEntry* altitudeEntry = yi4k ? exif_data_get_entry(data, exif_tag_from_name("GPSAltitude")) : NULL;
        ExifEntry* latitudeRefEntry = yi4k ? exif_data_get_entry(data, exif_tag_from_name("GPSLatitudeRef")) : NULL;
        ExifEntry* longitudeRefEntry = yi4k ? exif_data_get_entry(data, exif_tag_from_name("GPSLongitudeRef")) : NULL;
        ExifEntry* altitudeRefEntry = yi4k ? exif_data_get_entry(data, exif_tag_from_name("GPSAltitudeRef")) : NULL;

        /*if (!versionIdEntry) {
            printf("  Creating GPSVersionID entry\n");
            versionIdEntry = create_tag(data, EXIF_IFD_GPS, exif_tag_from_name("GPSVersionID"), 4, 4, EXIF_FORMAT_BYTE);
        }*/

        if (!latitudeEntry) {
            printf("  Creating GPSLatitude entry\n");
            latitudeEntry = create_tag(data, EXIF_IFD_GPS, exif_tag_from_name("GPSLatitude"), 24, 3, EXIF_FORMAT_RATIONAL);
        }
        if (!longitudeEntry) {
            printf("  Creating GPSLongitude entry\n");
            longitudeEntry = create_tag(data, EXIF_IFD_GPS, exif_tag_from_name("GPSLongitude"), 24, 3, EXIF_FORMAT_RATIONAL);
        }
        if (!altitudeEntry) {
            printf("  Creating GPSAltitude entry\n");
            altitudeEntry = create_tag(data, EXIF_IFD_GPS, exif_tag_from_name("GPSAltitude"), 8, 1, EXIF_FORMAT_RATIONAL);
        }

        if (!latitudeRefEntry) {
            printf("  Creating GPSLatitudeRef entry\n");
            latitudeRefEntry = create_tag(data, EXIF_IFD_GPS, exif_tag_from_name("GPSLatitudeRef"), 2, 2, EXIF_FORMAT_ASCII);
        }
        if (!longitudeRefEntry) {
            printf("  Creating GPSLongitudeRef entry\n");
            longitudeRefEntry = create_tag(data, EXIF_IFD_GPS, exif_tag_from_name("GPSLongitudeRef"), 2, 2, EXIF_FORMAT_ASCII);
        }
        if (!altitudeRefEntry) {
            printf("  Creating GPSAltitudeRef entry\n");
            altitudeRefEntry = create_tag(data, EXIF_IFD_GPS, exif_tag_from_name("GPSAltitudeRef"), 1, 1, EXIF_FORMAT_BYTE);
        }

        exif_set_rational(latitudeEntry->data, yi4k ? EXIF_BYTE_ORDER_INTEL : EXIF_BYTE_ORDER_MOTOROLA, { (unsigned)abs(ll.lat),  10000000 });
        exif_set_rational(latitudeEntry->data+8, yi4k ? EXIF_BYTE_ORDER_INTEL : EXIF_BYTE_ORDER_MOTOROLA, { 0, 1 });
        exif_set_rational(latitudeEntry->data+16, yi4k ? EXIF_BYTE_ORDER_INTEL : EXIF_BYTE_ORDER_MOTOROLA, { 0, 1 });
        exif_set_rational(longitudeEntry->data, yi4k ? EXIF_BYTE_ORDER_INTEL : EXIF_BYTE_ORDER_MOTOROLA, { (unsigned)abs(ll.lon), 10000000 });
        exif_set_rational(longitudeEntry->data+8, yi4k ? EXIF_BYTE_ORDER_INTEL : EXIF_BYTE_ORDER_MOTOROLA, { 0, 1 });
        exif_set_rational(longitudeEntry->data+16, yi4k ? EXIF_BYTE_ORDER_INTEL : EXIF_BYTE_ORDER_MOTOROLA, { 0, 1 });

        exif_set_rational(altitudeEntry->data, yi4k ? EXIF_BYTE_ORDER_INTEL : EXIF_BYTE_ORDER_MOTOROLA, { 20, 1 });

        ExifShort latRef = ll.lat >= 0 ? 'N' : 'S';
        ExifShort lonRef = ll.lon >= 0 ? 'E' : 'W';
        exif_set_short(latitudeRefEntry->data, yi4k ? EXIF_BYTE_ORDER_INTEL : EXIF_BYTE_ORDER_INTEL, latRef);
        exif_set_short(longitudeRefEntry->data, yi4k ? EXIF_BYTE_ORDER_INTEL : EXIF_BYTE_ORDER_INTEL, lonRef);
        //exif_set_short(altitudeRefEntry->data, EXIF_BYTE_ORDER_INTEL, 0); // above sea level

        /*exif_entry_dump(latitudeEntry,4);
        exif_entry_dump(latitudeRefEntry,4);
        exif_entry_dump(longitudeEntry,4);
        exif_entry_dump(longitudeRefEntry,4);
        exif_entry_dump(altitudeEntry,4);*/

        unsigned char* rawdata;
        unsigned int rawlen;
        exif_data_save_data(data, &rawdata, &rawlen);

        //printf("%d bytes in exif raw data\n", rawlen);

        FILE* fout = fopen(outfilename.c_str(),"wb");
        unsigned char hdr1[] = {0xFF,0xD8};
        fwrite(hdr1,1,2,fout);
        unsigned char hdr2[] = {0xFF,0xE1};
        fwrite(hdr2,1,2,fout);
        blockLen = rawlen + 2;
        fwrite(&swizzle[1],1,1,fout);
        fwrite(&swizzle[0],1,1,fout);
        fwrite(rawdata,rawlen,1,fout);

        //printf("fin is at %d\n",ftell(fin));

        char c;
        while ( fread(&c, 1, 1, fin) > 0 )
            fwrite(&c,1,1,fout);

        fclose(fout);
        fclose(fin);
    }

    return 0;
}
